from .base import (
    AbstractApplication,
    AbstractClient,
    BotContext,
    ReplyImage,
    ReplyTyping,
)
from .bot import GPTelBot

__all__ = [
    "GPTelBot",
    "AbstractApplication",
    "AbstractClient",
    "BotContext",
    "ReplyTyping",
    "ReplyImage",
]

__version__ = "0.2.0"



from abc import abstractmethod
from typing import Callable, Dict, Optional
from dataclasses import dataclass, field


@dataclass
class BotCommand:
    command: str
    description: str
    handler: Callable
    show_in_menu: Optional[bool] = True


@dataclass
class ApplicationConfig:
    name: str
    description: str
    token: str
    data_default: Dict = field(default_factory=dict)


class AbstractApplication:
    data_default: Dict = {}
    config: ApplicationConfig

    def __init__(self, config: ApplicationConfig):
        self.config = config

    @abstractmethod
    async def setup(self):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def add_handler(self, bot_command: BotCommand):
        pass

    @abstractmethod
    def set_chat_handler(self, handler: Callable):
        pass

    @abstractmethod
    def set_audio_handler(self, handler: Callable):
        pass


class AbstractClient:
    @abstractmethod
    def get_application(self, config: ApplicationConfig) -> AbstractApplication:
        pass


class ReplyBase:
    pass


class ReplyTyping(ReplyBase):
    pass


class ReplyImage(ReplyBase):
    url: str

    def __init__(self, url: str):
        self.url = url


class BotContext:
    def __init__(self):
        self.user_html = ""
        self.message = ""
        self.tmp_audio_path = None
        self.data = {}



import os
import tempfile
from contextlib import contextmanager


@contextmanager
def temporary_file_path(suffix=""):
    fd, path = tempfile.mkstemp(suffix=suffix)
    try:
        os.close(fd)
        yield path
    finally:
        os.remove(path)



import sys
import os
from typing import Callable, Dict, Optional, List
import asyncio
import argparse

from .base import AbstractClient, BotCommand, ApplicationConfig
from .telegram_client import TelegramClient


class GPTelBot:
    def __init__(
        self,
        name: str,
        description: str,
        data_default: Dict = {},
        token: Optional[str] = None,
        client: Optional[AbstractClient] = None,
    ):
        self.name = name
        self.description = description
        self.data_default = data_default
        self.token = token or os.getenv("TELEGRAM_TOKEN")
        if not self.token:
            raise ValueError("No token")
        self.client = client or TelegramClient()
        self.application = self.client.get_application(
            ApplicationConfig(
                name=self.name,
                description=self.description,
                token=self.token,
                data_default=self.data_default,
            )
        )
        self.bot_commands: List[BotCommand] = []
        self.chat_handler: Optional[Callable] = None
        self.audio_handler: Optional[Callable] = None

    def command(self, command: str, description: str):
        def decorator(func: Callable):
            self.bot_commands.append(
                BotCommand(command=command, description=description, handler=func)
            )
            return func

        return decorator

    def chat(self, audio: bool = True):
        def decorator(func: Callable):
            self.chat_handler = func
            if audio:
                self.audio_handler = func
            return func

        return decorator

    def audio(self):
        def decorator(func: Callable):
            self.audio_handler = func
            return func

        return decorator

    async def asetup(self):
        await self.application.setup()

    def setup(self):
        for command in self.bot_commands:
            self.application.add_handler(command)
        asyncio.run(self.asetup())

    def _parse_arguments(self, arguments):
        parser = argparse.ArgumentParser(description="GPTelBot")
        parser.add_argument("--setup", action="store_true", help="Setup bot")
        return parser.parse_args(arguments)

    def run(self):
        arguments = self._parse_arguments(sys.argv[1:])

        if arguments.setup:
            self.setup()
            return

        for bot_command in self.bot_commands:
            self.application.add_handler(bot_command)
        self.application.set_chat_handler(self.chat_handler)
        self.application.set_audio_handler(self.audio_handler)
        self.application.run()



from typing import Callable, Dict, List

from telegram import Update
from telegram.constants import ChatAction
from telegram.ext import (
    Application,
    CommandHandler,
    ContextTypes,
    MessageHandler,
    filters,
)

from .base import (
    AbstractApplication,
    AbstractClient,
    ApplicationConfig,
    BotContext,
    ReplyImage,
    ReplyTyping,
    BotCommand,
)
from .services import TranscriptionClient
from .utils import temporary_file_path


class TelegramBotContext(BotContext):
    def __init__(
        self,
        update: Update,
        context: ContextTypes.DEFAULT_TYPE,
        data_default: Dict = {},
    ):
        self.update = update
        self.context = context
        self.bot = context.bot
        self.args = context.args
        self.chat_id = update.effective_chat.id if update.effective_chat else None
        self.user = update.effective_user
        self.message = update.message

        self.user_html = self.user.mention_html()
        self.message = self.update.message.text
        self.tmp_audio_path = None
        self.data = context.chat_data  # A new dictionary should not be created (FIXME)
        for key, value in data_default.items():
            self.data.setdefault(key, value)


async def telegram_send_typing(update, context):
    await context.bot.send_chat_action(
        chat_id=update.effective_message.chat_id, action=ChatAction.TYPING
    )


def telegram_wrapper(handler: Callable, data_default: Dict = {}) -> Callable:
    async def wrapped_function(update: Update, context: ContextTypes.DEFAULT_TYPE):
        async_gen = handler(TelegramBotContext(update, context, data_default))
        async for message in async_gen:
            if isinstance(message, str):
                await update.message.reply_html(message)
            elif isinstance(message, ReplyTyping):
                await telegram_send_typing(update, context)
            elif isinstance(message, ReplyImage):
                await update.message.reply_photo(message.url)

    return wrapped_function


def telegram_text_wrapper(handler: Callable) -> Callable:
    async def wrapped_function(context: BotContext):
        async_gen = handler(context, context.message)
        async for message in async_gen:
            yield message

    return wrapped_function


def telegram_audio_wrapper(handler: Callable) -> Callable:
    async def wrapped_function(context: TelegramBotContext):
        yield ReplyTyping()
        with temporary_file_path(suffix=".ogg") as tmp_audio_path:
            context.tmp_audio_path = tmp_audio_path
            file = await context.context.bot.get_file(
                context.update.message.voice.file_id
            )
            await file.download_to_drive(tmp_audio_path)

            transcription_client = TranscriptionClient()
            context.message = await transcription_client.transcript(tmp_audio_path)

            async_gen = handler(context, context.message)
            async for message in async_gen:
                yield message

    return wrapped_function


class TelegramApplication(AbstractApplication):
    def __init__(self, config: ApplicationConfig):
        super().__init__(config)
        self.application = Application.builder().token(self.config.token).build()
        self.data_default = self.config.data_default
        self.commands_menu: List[BotCommand] = []

    async def setup(self):
        bot = self.application.bot
        await bot.set_my_name(self.config.name)
        await bot.set_my_description(self.config.description)
        await bot.set_my_commands(
            (
                (bot_command.command, bot_command.description)
                for bot_command in self.commands_menu
            )
        )

    def run(self):
        self.application.run_polling(allowed_updates=Update.ALL_TYPES)

    def add_handler(self, bot_command: BotCommand):
        self.application.add_handler(
            CommandHandler(
                bot_command.command,
                telegram_wrapper(bot_command.handler, self.data_default),
            )
        )
        self.commands_menu.append(bot_command)

    def set_chat_handler(self, handler: Callable):
        self.application.add_handler(
            MessageHandler(
                filters.TEXT & ~filters.COMMAND,
                telegram_wrapper(telegram_text_wrapper(handler), self.data_default),
            )
        )

    def set_audio_handler(self, handler: Callable):
        self.application.add_handler(
            MessageHandler(
                filters.VOICE,
                telegram_wrapper(telegram_audio_wrapper(handler), self.data_default),
            )
        )


class TelegramClient(AbstractClient):
    def __init__(self):
        pass

    def get_application(self, config: ApplicationConfig) -> AbstractApplication:
        return TelegramApplication(config)



import os

from openai import AsyncOpenAI


class TranscriptionClient:
    def __init__(self):
        self.client = AsyncOpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

    async def transcript(self, audio_path: str) -> str:
        audio_file = open(audio_path, "rb")
        transcription = await self.client.audio.transcriptions.create(
            model="whisper-1",
            file=audio_file,
            response_format="text",
        )

        return transcription
